// NeoGeo CD OBJ (de)compressor
// Converts standard NeoGeo CD SPR files to RLE compressed OBJ and vice versa
// based on information from: https://wiki.neogeodev.org/index.php?title=OBJ_file
// January 2025, IQ_132, neo-source.com

// Note: some data in obj files may be junk and is not re-created when building obj files

#include <stdio.h>

// endian swap
#define SWAP(x)	(((x & 0xff) << 24) | ((x & 0xff00) << 8) | ((x >> 8) & 0xff00) | ((x >> 24) & 0xff))

static unsigned int decompress(unsigned char *src, unsigned char *dst, unsigned int length)
{
	int i;
	int ptr_offs = 0;
	int out_offs = 0;
	unsigned int *ptr = (unsigned int*)src;
	unsigned int *out = (unsigned int*)dst;

	for (i = 0; i < (length / 4); i += (0x800/4)) // sectors are 2048 bytes in size
	{
		ptr_offs = i; // sector align

		// how many copies or zero-fill commands to process
		unsigned int process_count = SWAP(ptr[ptr_offs]); ptr_offs++;

		// valid process_count is always has top word 0x0001
		if ((process_count & 0xffff0000) == 0x00010000)
		{
			process_count &= 0xffff; // trim off top word

			while (process_count != -1)
			{
				unsigned int tmp = ptr[ptr_offs++]; // get next long and check for zero value

				if (tmp == 0) // data is a zero, use a "fill" command
				{
					// the next long after a zero is how many zeros to copy into the destination
					unsigned int zero_count = SWAP(ptr[ptr_offs]) & 0xffff; ptr_offs++;

					while (zero_count != -1)
					{
						out[out_offs++] = 0;
						zero_count--;
					}

					process_count--;
				}
				else // direct copy
				{
					out[out_offs++] = tmp;
					process_count--;
				}

				// ensure output stays inside of allocated memory
				if (out_offs >= (0x100000/4))
				{
					i = 0x100000;
					break;
				}
			}
		}
	}
	
	return (out_offs * 4); // length of decompressed file
}

static unsigned int compress(unsigned char *src, unsigned char *dst, unsigned int length)
{
	int ptr_offs = 0; // input file position
	int out_offs = 0; // output file position
	int op_counter = 0; // count how many operations per sector
	unsigned int *ptr = (unsigned int *)src;
	unsigned int *out = (unsigned int *)dst;

	while (ptr_offs < length/4)
	{
		// first long word in each sector is how many copies / zero-fills that are to be performed (n-1 due to dbra/dbf)
		if ((out_offs & 0x1ff) == 0)
		{
			op_counter = 0;
			out[out_offs++] = SWAP(0x0000ffff); // result is always 0x10000 + n - 1
		}

		if (ptr[ptr_offs] == 0)	// this long word is 0!
		{
			// 0-fill command takes two long words, and they must both be within the same sector!
			if ((out_offs & 0x1ff) == 0x1ff) {
				out_offs++; // fill with junk?
				continue;
			}

			out[out_offs++] = 0;	// set value to zero, this triggers a 'fill' command
			ptr_offs++;

			unsigned int zero_count = 0; // should be 1, but 'fill' command is n-1 (dbra/dbf)

			while ((ptr[ptr_offs] == 0) && (ptr_offs < length/4) && (op_counter < 0x9c4)) // go until a non-zero value is found, eof, or 9c4 total copies
			{
				ptr_offs++;
				zero_count++;
				op_counter++;
			}

			out[out_offs & ~0x1ff] = SWAP(SWAP(out[out_offs & ~0x1ff]) + 1); // increment the first long word in this sector
			out[out_offs++] = SWAP(zero_count); // 'fill' command
			op_counter++;

			// from Alpha Mission II's obj files
			if (op_counter >= 0x9c4) { // if the total number of copies hits 0x9c4, stop here for this sector - number before the hardware hangs/crashes?
				out_offs = ((out_offs - 1) | 0x1ff) + 1;
			}
		}
		else	// everything else
		{
			out[out_offs & ~0x1ff] = SWAP(SWAP(out[out_offs & ~0x1ff]) + 1); // increment the first long word in this sector
			out[out_offs++] = ptr[ptr_offs++]; // copy
			op_counter++;
		}
	}
	
	return ((out_offs | 0x1ff) + 1) * 4; // length of compressed file (must be full sector size)
}

int main(int argc, char *argv[])
{
	if (argc != 3)
	{
		printf ("Usage:\nbuildobj (C)ompress/(D)ecompress filename\nExample:\n\tbuildobj c spr.spr\n\tbuildobj d spr.obj\n");
		return 1;
	}

	FILE *fa;
	int length = 0;
	char name_out[128];
	int mode = (tolower(argv[1][0]) == 'c') ? 1 : 0;

	printf ("Attempting to %scompress %s...\n", mode ? "" : "de", argv[2]);

	fa = fopen(argv[2], "rb");

	if (fa == NULL)
	{
		printf ("Unable to open %s\n", argv[1]);
		return 1;
	}

	fseek (fa, 0, SEEK_END);
	length = ftell (fa);

	if (length == 0)
	{
		printf ("Input file %s has 0 length!\n");
		return 1;
	}

	if (length > 0x100000)
	{
		printf ("Input file must be less than 0x100000 bytes (1MB) in length!\n");
		return 1;
	}

	rewind (fa);

	unsigned char *src = (unsigned char*)calloc(1,0x100000);
	unsigned char *dst = (unsigned char*)calloc(1,0x200000); // double size for crash prevention

	fread (src, length, 1, fa);
	fclose (fa);

	length = (mode) ? compress(src, dst, length) : decompress(src, dst, length);

	if (argv[2][strlen(argv[2])-4] == '.') argv[2][strlen(argv[2])-4] = 0; // remove filename extension

	if (length >= 0x100000 && mode != 0)
	{
		printf ("ERROR: Output file %s.OBJ not created.\n       Output file larger than input file.\n", argv[2]);
		free (dst);
		free (src);
		return 1;
	}

	sprintf (name_out, "%s.%s", argv[2], mode ? "OBJ" : "SPR");

	fa = fopen(name_out, "wb");
	fwrite (dst, length, 1, fa);
	fclose (fa);

	free (dst);
	free (src);

	printf ("Completed! %s created!\n", name_out);

	return 0;
}
